import java.awt.*;
import java.util.*;
import java.awt.image.BufferedImage;
import java.io.*;

/**
  *
  * Objekte der Klasse stellen einen virtuellen Hauptspeicher dar.
  *
  * @version 2.1 vom 05.01.2014
  * @author Daniel Garmann
  */

public class Hauptspeicher {
  
  // Anfang Attribute
  private final static String lineSeparator = System.getProperty("line.separator"); // Fr Dateiausgabe
  public final static int FREIEDATENSTRUKTUR = 0;
  public final static int LINEARELISTE = 1;
  public final static int STAPEL = 2;
  public final static int SCHLANGE = 3;
  
  private Graphics2D ausgabe;                                        // Zeichenflche
  private int groesseX, groesseY;                                    // Gre der Zeichenflche
  private ArrayList<Zeiger> zeigerSammlung = new ArrayList();        // Sammlung aller Zeiger im Hauptspeicher
  private ArrayList<Listenobjekt> objekteSammlung = new ArrayList(); // Sammlung aller Listenobjekte im Hauptspeicher
  private ArrayList<Punkt> punkteSammlung = new ArrayList();         // Sammlung aller Punkte (Endpunkt der Referenzen)
  
  private int listenstrukturtyp = FREIEDATENSTRUKTUR;                // Typ der Listenstruktur (Schlange, Stapel, Liste,...)
  
  private BufferedImage bufferedImage; // Double-Buffering (auf diesem Bild wird intern gemalt und dann ausgegeben
  private Graphics2D g;                // Graphics-Komponente des Bildes
  
  public Optionen optionen;

  // Ende Attribute
  
  // Anfang Methoden
  
  /**
   * Konstruktor fr einen neuen Hauptspeicher
   * @param pAusgabe Zeichenflche fr den Hauptspeicher
   * @param pGroesseX Breite der Zeichenflche
   * @param pGroesseY Hhe der Zeichenflche
   */
  public Hauptspeicher(Graphics2D pAusgabe, int pGroesseX, int pGroesseY) {
    ausgabe = pAusgabe;     
    groesseX = pGroesseX;
    groesseY = pGroesseY;
    bufferedImage = new BufferedImage(groesseX, groesseY,java.awt.image.BufferedImage.TYPE_INT_RGB);
    g = (Graphics2D)bufferedImage.getGraphics();
    optionen = new Optionen();
  }


  /**
   * Methode zum ndern der Ausgabegre
   * @param pGroesseX Breite der Zeichenflche
   * @param pGroesseY Hhe der Zeichenflche
   */
  public void setGroesse(Graphics2D pAusgabe, int pGroesseX, int pGroesseY) {
    ausgabe = pAusgabe;
    groesseX = pGroesseX;
    groesseY = pGroesseY;
    bufferedImage = new BufferedImage(groesseX, groesseY,java.awt.image.BufferedImage.TYPE_INT_RGB);
    g = (Graphics2D)bufferedImage.getGraphics();
  }
    
  /**
   * Methode zum Lschen aller Objekt im virtuellen Hauptspeicher
   */
  public void clear() {
    zeigerSammlung.clear();
    objekteSammlung.clear();
    punkteSammlung.clear();
  }
  
  /**
   * Methode zum Setzen des Typs der Listenstruktur, die dargestellt werden soll
   * @param typ neuer Listenstrukturtyp
   */
  public void setListenstrukturTyp(int typ) {
    clear();
    Zeiger z1, z2, z3;
    switch (typ) {
      case LINEARELISTE: 
             z1 = neuerZeiger(optionen.labelReferenceListFirst);
             z1.setStatisch(true);
             z2 = neuerZeiger(optionen.labelReferenceListCurrent);
             z2.setStatisch(true);
             if (optionen.fixedListReferences == 3) {
               z3 = neuerZeiger(optionen.labelReferenceListLast);
               z3.setStatisch(true);
             }
             break;
      case SCHLANGE: 
             z1 = neuerZeiger(optionen.labelReferenceQueueFirst);
             z1.setStatisch(true);
             z2 = neuerZeiger(optionen.labelReferenceQueueLast);
             z2.setStatisch(true);
             break;
      case STAPEL: 
             z1 = neuerZeiger(optionen.labelReferenceStackFirst);
             z1.setStatisch(true);
             break;
    } 
    this.listenstrukturtyp = typ;
  }
  
  /**
   * Methode zur Ermittlung des Listenstrukturtyps.
   * @return Listenstrukturtyp: 
   */
  public int gibListenstrukturTyp() {
    return listenstrukturtyp;
  }
   
  /**
   * Methode zum Speichern des Hauptspeicherinhalts in eine Datei
   * @param dateiname Der Dateiname der zu speichernden Datei
   */
  public void speichern(String dateiname) {
    try {
      FileWriter w = new FileWriter(dateiname);
      // Nacheinander zunchst alle Informationen der Listenobjekte speichern
      for (Listenobjekt o: objekteSammlung) {      
        w.write(""+o+","+o.getInhalt()+","+o.getX()+","+o.getY()+","+o.getReferenz()+lineSeparator);
      }  
      // dann alle Informationen der Zeiger speichern.
      for (Zeiger z: zeigerSammlung) {
        w.write(""+z+","+z.getName()+","+z.getX()+","+z.getY()+","+z.getReferenz()+","+z.istStatisch()+lineSeparator);
      }  
      w.close();      
    } catch (IOException e) {
      System.err.println(e.toString());
    }
  }
  
  /**
   * Methode zum ffnen des Hauptspeicherinhalts aus einer Datei
   * @param dateiname Der Dateiname der zu speichernden Datei
   */
  public void laden(String dateiname) {
    // Zunchst alle Objekte im Hauptspeicher lschen.
    clear();
    // Alle Objekte sind mit einem Schlssel abgespeichert, ber den die 
    // Objekte miteinander in Beziehung gesetzt werden knnen.
    // Die Objekte werden zunchst in eine Hashtabelle eingefgt, um die 
    // tatschlichen Referenzierungen im Anschluss herzustellen.
    Hashtable<String,Zeiger> zeigerHashliste = new Hashtable();
    Hashtable<String,Listenobjekt> objekteHashliste = new Hashtable();
    
    try {
      FileReader r = new FileReader(dateiname);
      BufferedReader br = new BufferedReader(r);
      String s = br.readLine();
      // Nacheinander alle Zeilen der Textdatei einlesen
      while (s != null) {
        // Jede Zeile enthlt einzelne, kommagetrennte Informationen zu dem Hauptspeicherobjekt
        String[] teile = s.split(",");
        // bei einem Zeiger sind die Informtionen:
        // Hash-Schlssel, Zeigername, x-Position, y-Position, Refernz-Hash-Schlssel
        if (s.startsWith("Zeiger")) {
          Zeiger z = new Zeiger(Integer.parseInt(teile[2]),Integer.parseInt(teile[3]),teile[1],null, optionen);
          if (teile.length >= 6 && teile[5].equals("true")) {
            z.setStatisch(true);
          };
          punkteSammlung.add(z.getZeigerEnde());
          zeigerHashliste.put(teile[0],z);
        // bei einem Listenobjekt sind die Informtionen:
        // Hash-Schlssel, Inhalt des Listenobjekts, x-Position, y-Position, Refernz-Hash-Schlssel
        } else if (s.startsWith("Listenobjekt")) {
          Listenobjekt o = new Listenobjekt(Integer.parseInt(teile[2]),Integer.parseInt(teile[3]),teile[1],null, optionen);
          objekteHashliste.put(teile[0],o);
          punkteSammlung.add(o.getZeigerEnde());
        }
        s = br.readLine();   
      }
      r.close();

      // In einem Zweiten Durchlauf knnen nun die Referenzen gesetzt werden und
      // die Objekte im Hauptspeicher hinzugefgt werden
      r = new FileReader(dateiname);
      br = new BufferedReader(r);
      s = br.readLine();
      while (s != null) {
        String[] teile = s.split(",");        
        if (s.startsWith("Zeiger")) {
          Listenobjekt referenz = objekteHashliste.get(teile[4]);
          Zeiger z = zeigerHashliste.get(teile[0]);
          z.setReferenz(referenz);
        } else if (s.startsWith("Listenobjekt")) {
          Listenobjekt referenz = objekteHashliste.get(teile[4]);
          Listenobjekt o = objekteHashliste.get(teile[0]);
          o.setReferenz(referenz);
        }
        s = br.readLine();   
      }
      r.close();

      // Nachdem alle Referenzen gesetzt wurde, knnen alle Zeiger und
      // alle Listenobjekte der Hash-Tabelle in die Sammlung der Zeiger
      // und Listenobjekte bernommen werden
      zeigerSammlung.addAll(zeigerHashliste.values());
      objekteSammlung.addAll(objekteHashliste.values());
      
      // Die Punkte der Zeiger-Enden mssen noch aktualisiert werden
      aktualisiereListenobjekte();
      
    } catch (IOException e) {
      System.err.println(e.toString());
    }
  }
  
  /**
   * Methode zur Darstellung des virtuellen Hauptspeichers
   */
  public void darstellen() {
    // Nur, wenn eine Zeichenflche exisitert, dann darf gezeichnet werden
    if (ausgabe != null) {
      // Double-Buffering, also zweites Bild, auf dem gezeichnet wird, und
      // dann erst ganz zum Schluss darstellen
      BufferedImage bildImage = new BufferedImage(groesseX, groesseY,java.awt.image.BufferedImage.TYPE_INT_RGB);
      Graphics2D g = (Graphics2D)bildImage.getGraphics();
  
      // Bild lschen
      g.setColor(Color.LIGHT_GRAY);
      g.fillRect(0, 0, groesseX, groesseY);
  
      // Alle Referenzen auf ein Objekt sollen untereinander dargestellt werden.
      for (Listenobjekt aktuellesObjekt: objekteSammlung) {
        int i = 0;
        for (Punkt aktuellerPunkt: punkteSammlung) {
          if (aktuellerPunkt.getBesitzer().getReferenz() == aktuellesObjekt) {
            aktuellerPunkt.setX(aktuellesObjekt.getX()-5);
            int y = 25 + ((i % 2)*2-1) * ((i+1) / 2) * 10;
            aktuellerPunkt.setY(aktuellesObjekt.getY()+y);
            i++;
          }
        }
      }    
  
      // Darstellung der Zeiger
      for (Zeiger aktuellerZeiger: zeigerSammlung) {
        aktuellerZeiger.darstellen(g);        
      }
      // Darstellung der Listenobjekte
      for (Listenobjekt aktuellesObjekt: objekteSammlung) {
        aktuellesObjekt.darstellen(g);
      }    
      // Darstellung der Punkte fr die Zeiger-Enden
      for (Punkt aktuellerPunkt: punkteSammlung) {
        aktuellerPunkt.darstellen(g);
      }    
      ausgabe.drawImage(bildImage,0,0,null);
    }
  }
    
  /**
   * Methode zur Ermittlung des Objekts, welches sich am Punkt (x|y) befindet
   * @param x x-Koordinate auf dem virtuellen Hauptspeicher
   * @param y y-Koordinate auf dem virtuellen Hauptspeicher
   * @return das Hauptspeicherobjekt, welches sich an der Stelle (x|y) befindet.
   */
  public Darstellungsobjekt gibObjektAnStelle(int x, int y) {
    Darstellungsobjekt result = null;
    for (Punkt aktuellerPunkt: punkteSammlung) {
      if (x >= aktuellerPunkt.getX()-5 && x <= aktuellerPunkt.getX() + 5 &&
          y >= aktuellerPunkt.getY()-5 && y <= aktuellerPunkt.getY() + 5) {
        result = aktuellerPunkt;
      }      
    }
    for (Zeiger aktuellerZeiger: zeigerSammlung) {
      if (x >= aktuellerZeiger.getX() && x <= aktuellerZeiger.getX() + 50 &&
          y >= aktuellerZeiger.getY() && y <= aktuellerZeiger.getY() + 50) {
        result = aktuellerZeiger;
      }      
    }
    for (Listenobjekt aktuellesObjekt: objekteSammlung) {
      if (x >= aktuellesObjekt.getX() && x <= aktuellesObjekt.getX() + 100 &&
          y >= aktuellesObjekt.getY() && y <= aktuellesObjekt.getY() + 50) {
        result = aktuellesObjekt;
      }      
    }
    return result;
  }
  
  
  /**
   * Methode zur berprfung, ob ein Hauptspeicherobjekt ein anderes Objekt berdeckt.
   * @param objekt Das Objekt, von dem berprft wird, ob es auf einem anderen liegt.
   * @return true, falls das Objekt ein anderes berdeckt, sonst false
   */
  public boolean ueberdecktHauptspeicherobjekt(Hauptspeicherobjekt objekt) {
    Darstellungsobjekt o1, o2, o3, o4, o5;
    o1 = gibObjektAnStelle(objekt.getX(),objekt.getY());
    o2 = gibObjektAnStelle(objekt.getX(),objekt.getY()+50);    
    if (objekt instanceof Zeiger) {
      o3 = gibObjektAnStelle(objekt.getX()+50,objekt.getY());
      o4 = gibObjektAnStelle(objekt.getX()+50,objekt.getY()+50);    
      o5 = gibObjektAnStelle(objekt.getX()+25,objekt.getY()+25);    
    } else {
      o3 = gibObjektAnStelle(objekt.getX()+100,objekt.getY());
      o4 = gibObjektAnStelle(objekt.getX()+100,objekt.getY()+50);    
      o5 = gibObjektAnStelle(objekt.getX()+50,objekt.getY()+25);    
    }
    
    return ( (o1 != null && o1 != objekt) || 
             (o2 != null && o2 != objekt) || 
             (o3 != null && o3 != objekt) || 
             (o4 != null && o4 != objekt) ||
             (o5 != null && o5 != objekt) );    
  } 
   
  /**
   * Methode zur berprfung, ob ein Zeigername bereits vergeben ist
   * @param name der Zeigername, der berprft werden soll
   * @return true, falls der Zeigername bereits exisitiert, sonst false
   */
  public boolean exisitiertZeiger(String name) {
    Zeiger zeiger = null;
    for (Zeiger aktuellerZeiger: zeigerSammlung) {
      if (aktuellerZeiger.getName().equals(name)) {
        return true;
      } 
    }
    return false;
  }
  
  /**
   * Methode zur Ermittlung, wie viele Zeiger bereits erstellt wurden.
   * @return Anzahl der erstellten Zeiger
   */
  public int gibAnzahlZeiger() {
    return zeigerSammlung.size();
  } 
  
  /**
   * Methode zur Erstellung eines neuen Zeigers
   * @param name der Name des neuen Zeigers, der erstellt werden soll
   * @return ein neues Zeiger-Objekt
   */
  public Zeiger neuerZeiger(String name) {  
    if (exisitiertZeiger(name) == false) {      
      // Neuen Zeiger unter die bereits bestehenden anfgen
      Zeiger zeiger = new Zeiger(50,50+zeigerSammlung.size()*100,name,null, optionen);
      
      while (zeiger.getY() > groesseY - 100 || ueberdecktHauptspeicherobjekt(zeiger)) {
        zeiger.setX((int)(Math.random()*(groesseX / 2)));
        zeiger.setY((int)(Math.random()*(groesseY-100)));
      }
      zeigerSammlung.add(zeiger);
      Punkt zeigerEnde = zeiger.getZeigerEnde();
      zeigerEnde.setX(zeiger.getX()+80);
      zeigerEnde.setY(zeiger.getY()+25);
      punkteSammlung.add(zeigerEnde);
      return zeiger;
    }
    return null;
  }

  /**
   * Methode zur Lschung eines Zeigers aus dem Hauptspeicher
   * @param name der Name des zu lschenden Zeigers
   */
  public void loescheZeiger(Zeiger zeiger) {
    if (zeiger != null) {
      punkteSammlung.remove(zeiger.getZeigerEnde());
      zeigerSammlung.remove(zeiger);
    }
  }
  
  /**
   * Methode zur Erstellung eines neuen Listenobjekts
   * @param zeiger der Zeiger, an dem das neue Objekt erstellt werden soll
   * @return ein neues Listenobjekt
   */
  public Listenobjekt neuesObjekt(Zeiger zeiger) {
    Listenobjekt listenObjekt = new Listenobjekt(zeiger,"",null, optionen);
    int setzversuche = 0;
    while (ueberdecktHauptspeicherobjekt(listenObjekt) && setzversuche < 10000) {
      listenObjekt.setX((int)(Math.random()*(groesseX-150)));
      listenObjekt.setY((int)(Math.random()*(groesseY-100)));
      setzversuche++;
    }
    objekteSammlung.add(listenObjekt);
    Punkt zeigerEnde = listenObjekt.getZeigerEnde();
    zeigerEnde.setX(listenObjekt.getX()+130);
    zeigerEnde.setY(listenObjekt.getY()+25);
    punkteSammlung.add(zeigerEnde);
    return listenObjekt;
  } 

  /**
   * Methode zur Reaktivierung eines Listenobjekts
   * @param objekte ArrayList mit den zu aktivierenden Darstellungsobjekten
   */
  public void reaktivereDarstellungsobjekte(ArrayList<Darstellungsobjekt> objekte) {
    for (Darstellungsobjekt objekt: objekte) {
      if (objekt instanceof Punkt) {
        punkteSammlung.add((Punkt)objekt);
      } else if (objekt instanceof Listenobjekt) {
        objekteSammlung.add((Listenobjekt)objekt);
      } else if (objekt instanceof Zeiger) {
        zeigerSammlung.add((Zeiger)objekt);
      }
    }
  } 
  
  /**
   * Methode zur Lschung eines Listenobjekts aus dem Hauptspeicher
   * @param objekt das zu lschende Objekt
   */
  public void loescheObjekt(Listenobjekt objekt) {
    if (objekt != null) {
      punkteSammlung.remove(objekt.getZeigerEnde());
      objekteSammlung.remove(objekt);
    } 
  }
  
  /**
   * Methode zur Aktualisierung aller Listenobjekte. Hier wird festgelegt,
   * ob die Listenobjekte ber einen Zeiger direkt erreichbar (DIREKTERREICHBAR),
   * ber ein Listenobjekt in zweiter Reihe erreichbar (INDIREKTERREICHBAR),
   * berhaupt erreichbar ber andere Listenobjekte (ERREICHBAR),
   * oder gar nicht erreichbar sind (NICHTERREICHBAR)
   */
  public void aktualisiereListenobjekte() {
    // Zunchst alle Listenobjekte als nicht erreichbar setzen.
    for (Listenobjekt aktuellesObjekt: objekteSammlung) {
      aktuellesObjekt.setStatus(Listenobjekt.NICHTERREICHBAR);
      // Position der Zeiger-Ende-Punkte aktualisieren
      aktuellesObjekt.setX(aktuellesObjekt.getX());
      aktuellesObjekt.setY(aktuellesObjekt.getY());        
    } 
    // Dann fr alle Zeiger nachsehen, welche Elemente von dem Zeiger aus erreichbar sind.   
    for (Zeiger aktuellerZeiger: zeigerSammlung) {
      aktuellerZeiger.setX(aktuellerZeiger.getX());
      aktuellerZeiger.setY(aktuellerZeiger.getY());        
      Listenobjekt l = aktuellerZeiger.getReferenz();
      if (l != null) {
        l.setStatus(Listenobjekt.DIREKTERREICHBAR);
        if (l.getReferenz() != null && l.getReferenz().getStatus() > Listenobjekt.INDIREKTERREICHBAR) {
          l = l.getReferenz();
          l.setStatus(Listenobjekt.INDIREKTERREICHBAR);
        }        
        while (l.getReferenz() != null && l.getReferenz().getStatus() == Listenobjekt.NICHTERREICHBAR) {
          l = l.getReferenz();
          l.setStatus(Listenobjekt.ERREICHBAR);
        }
      }                
    }
  }
  
  /**
   * Methode zur Lschung aller nicht erreichbarer Listenobjekte
   */
  public ArrayList<Darstellungsobjekt> speicherbereinigung() {
    ArrayList<Darstellungsobjekt> geloeschteObjekte = new ArrayList();
    
    Object[] objekte = objekteSammlung.toArray();
    for (int i = 0; i < objekte.length; i++) {
      Listenobjekt aktuellesObjekt = (Listenobjekt)objekte[i];
      if (aktuellesObjekt.getStatus() == Listenobjekt.NICHTERREICHBAR) {
        geloeschteObjekte.add(aktuellesObjekt.getZeigerEnde());
        geloeschteObjekte.add(aktuellesObjekt);
        
        punkteSammlung.remove(aktuellesObjekt.getZeigerEnde());
        objekteSammlung.remove(aktuellesObjekt);        
      }
    }
    return geloeschteObjekte;  
  }  
  
  // Ende Methoden
}
